#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Example Airflow DAG to submit Apache Spark applications using
`SparkSubmitOperator`, `SparkJDBCOperator` and `SparkSqlOperator`.
"""
from airflow.models import DAG
from airflow.providers.apache.drill.operators.drill import DrillOperator
from airflow.utils.dates import days_ago

args = {
    'owner': 'Airflow',
}

with DAG(
    dag_id='example_drill_dag',
    default_args=args,
    schedule_interval=None,
    start_date=days_ago(2),
    tags=['example'],
) as dag:
    # [START howto_operator_drill]
    sql_task = DrillOperator(
        task_id='json_to_parquet_table',
        sql='''
        drop table if exists dfs.tmp.employee;
        create table dfs.tmp.employee as select * from cp.`employee.json`;
        ''',
    )
    # [END howto_operator_drill]



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from typing import Iterable, Mapping, Optional, Union

import sqlparse

from airflow.models import BaseOperator
from airflow.providers.apache.drill.hooks.drill import DrillHook
from airflow.utils.decorators import apply_defaults


class DrillOperator(BaseOperator):
    """
    Executes the provided SQL in the identified Drill environment.

    .. seealso::
        For more information on how to use this operator, take a look at the guide:
        :ref:`howto/operator:DrillOperator`

    :param sql: the SQL code to be executed. (templated)
    :type sql: Can receive a str representing a sql statement,
        a list of str (sql statements), or a reference to a template file.
        Template references are recognized by str ending in '.sql'
    :param drill_conn_id: id of the connection config for the target Drill
        environment
    :type drill_conn_id: str
    :param parameters: (optional) the parameters to render the SQL query with.
    :type parameters: dict or iterable
    """

    template_fields = ('sql',)
    template_fields_renderers = {'sql': 'sql'}
    template_ext = ('.sql',)
    ui_color = '#ededed'

    @apply_defaults
    def __init__(
        self,
        *,
        sql: str,
        drill_conn_id: str = 'drill_default',
        parameters: Optional[Union[Mapping, Iterable]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.sql = sql
        self.drill_conn_id = drill_conn_id
        self.parameters = parameters
        self.hook = None

    def execute(self, context):
        self.log.info('Executing: %s on %s', self.sql, self.drill_conn_id)
        self.hook = DrillHook(drill_conn_id=self.drill_conn_id)
        sql = sqlparse.split(sqlparse.format(self.sql, strip_comments=True))
        no_term_sql = [s[:-1] for s in sql if s[-1] == ';']
        self.hook.run(no_term_sql, parameters=self.parameters)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from typing import Any, Iterable, Optional, Tuple

from sqlalchemy import create_engine
from sqlalchemy.engine import Connection

from airflow.hooks.dbapi import DbApiHook


class DrillHook(DbApiHook):
    """
    Interact with Apache Drill via sqlalchemy-drill.

    You can specify the SQLAlchemy dialect and driver that sqlalchemy-drill
    will employ to communicate with Drill in the extras field of your
    connection, e.g. ``{"dialect_driver": "drill+sadrill"}`` for communication
    over Drill's REST API.  See the sqlalchemy-drill documentation for
    descriptions of the supported dialects and drivers.

    You can specify the default storage_plugin for the sqlalchemy-drill
    connection using the extras field e.g. ``{"storage_plugin": "dfs"}``.
    """

    conn_name_attr = 'drill_conn_id'
    default_conn_name = 'drill_default'
    conn_type = 'drill'
    hook_name = 'Drill'
    supports_autocommit = False

    def get_conn(self) -> Connection:
        """Establish a connection to Drillbit."""
        conn_md = self.get_connection(getattr(self, self.conn_name_attr))
        creds = f'{conn_md.login}:{conn_md.password}@' if conn_md.login else ''
        engine = create_engine(
            f'{conn_md.extra_dejson.get("dialect_driver", "drill+sadrill")}://{creds}'
            f'{conn_md.host}:{conn_md.port}/'
            f'{conn_md.extra_dejson.get("storage_plugin", "dfs")}'
        )

        self.log.info(
            'Connected to the Drillbit at %s:%s as user %s', conn_md.host, conn_md.port, conn_md.login
        )
        return engine.raw_connection()

    def get_uri(self) -> str:
        """
        Returns the connection URI

        e.g: ``drill://localhost:8047/dfs``
        """
        conn_md = self.get_connection(getattr(self, self.conn_name_attr))
        host = conn_md.host
        if conn_md.port is not None:
            host += f':{conn_md.port}'
        conn_type = 'drill' if not conn_md.conn_type else conn_md.conn_type
        dialect_driver = conn_md.extra_dejson.get('dialect_driver', 'drill+sadrill')
        storage_plugin = conn_md.extra_dejson.get('storage_plugin', 'dfs')
        return f'{conn_type}://{host}/{storage_plugin}' f'?dialect_driver={dialect_driver}'

    def set_autocommit(self, conn: Connection, autocommit: bool) -> NotImplemented:
        raise NotImplementedError("There are no transactions in Drill.")

    def insert_rows(
        self,
        table: str,
        rows: Iterable[Tuple[str]],
        target_fields: Optional[Iterable[str]] = None,
        commit_every: int = 1000,
        replace: bool = False,
        **kwargs: Any,
    ) -> NotImplemented:
        raise NotImplementedError("There is no INSERT statement in Drill.")
